package ru.dbotthepony.kstarbound.io.json

import com.google.common.collect.ImmutableList
import com.google.common.collect.ImmutableMap
import com.google.gson.JsonArray
import com.google.gson.JsonObject
import com.google.gson.JsonParseException
import com.google.gson.JsonSyntaxException
import com.google.gson.TypeAdapter
import com.google.gson.internal.bind.JsonTreeReader
import com.google.gson.internal.bind.TypeAdapters
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonToken
import com.google.gson.stream.JsonWriter
import it.unimi.dsi.fastutil.objects.Object2IntArrayMap
import it.unimi.dsi.fastutil.objects.ObjectArraySet
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.enrollList
import ru.dbotthepony.kstarbound.defs.enrollMap
import ru.dbotthepony.kstarbound.defs.flattenJsonElement
import ru.dbotthepony.kstarbound.getValue
import ru.dbotthepony.kstarbound.setValue
import java.lang.reflect.Constructor
import kotlin.jvm.internal.DefaultConstructorMarker
import kotlin.properties.Delegates
import kotlin.reflect.*
import kotlin.reflect.full.isSubclassOf
import kotlin.reflect.full.isSupertypeOf

/**
 * [TypeAdapter] для классов, которые имеют все свои свойства в главном конструкторе.
 */
class FactoryAdapter<T : Any> private constructor(
	val bound: KClass<T>,
	private val types: ImmutableList<PackedProperty<T, *>>,
	val asJsonArray: Boolean,
	val storesJson: Boolean,
	val logMisses: Boolean,
) : TypeAdapter<T>() {
	private data class PackedProperty<Clazz : Any, T>(
		val property: KProperty1<Clazz, T>,
		val adapter: TypeAdapter<T>,
		val transformer: (T) -> T = { it },
		val isFlat: Boolean
	) {
		val returnType = property.returnType
	}

	private val name2index = Object2IntArrayMap<String>()
	private val loggedMisses = ObjectArraySet<String>()

	init {
		name2index.defaultReturnValue(-1)

		for ((i, pair) in types.withIndex()) {
			name2index[pair.property.name] = i
		}
	}

	/**
	 * Обычный конструктор класса (без флагов "значения по умолчанию")
	 */
	private val regularFactory: KFunction<T> = bound.constructors.firstOrNull first@{
		var requiredSize = types.size

		if (storesJson)
			requiredSize++

		if (it.parameters.size == requiredSize) {
			val iterator = types.iterator()
			val factoryIterator = it.parameters.iterator()

			while (factoryIterator.hasNext() && iterator.hasNext()) {
				val param = factoryIterator.next()
				val nextParam = iterator.next()

				val a = param.type
				val b = nextParam.returnType

				if (!a.isSupertypeOf(b) || a.isMarkedNullable != b.isMarkedNullable) {
					return@first false
				}
			}

			if (storesJson) {
				val nextParam = factoryIterator.next()

				if (asJsonArray) {
					if (!(nextParam.type.classifier as KClass<*>).isSubclassOf(List::class)) {
						return@first false
					}
				} else {
					if (!(nextParam.type.classifier as KClass<*>).isSubclassOf(Map::class)) {
						return@first false
					}
				}
			}

			return@first true
		}

		return@first false
	} ?: throw NoSuchElementException("Unable to determine constructor for ${bound.qualifiedName} matching (${types.joinToString(", ")})")

	/**
	 * Синтетический конструктор класса, который создаётся Kotlin'ном, для создания классов со значениями по умолчанию
	 */
	private val syntheticFactory: Constructor<T>? = try {
		val typelist = types.map { (it.returnType.classifier as KClass<*>).java }.toMutableList()

		if (storesJson)
			if (asJsonArray)
				typelist.add(List::class.java)
			else
				typelist.add(Map::class.java)

		for (i in 0 until (if (types.size % 31 == 0) types.size / 31 else types.size / 31 + 1))
			typelist.add(Int::class.java)

		typelist.add(DefaultConstructorMarker::class.java)

		bound.java.getDeclaredConstructor(*typelist.toTypedArray())
	} catch(_: NoSuchMethodException) {
		null
	}

	private val syntheticPrimitives: Array<Any?>?

	init {
		if (syntheticFactory == null) {
			syntheticPrimitives = null
		} else {
			syntheticPrimitives = arrayOfNulls(syntheticFactory.parameters.size)

			for ((i, param) in syntheticFactory.parameters.withIndex()) {
				val type = param.parameterizedType as? Class<*> ?: continue

				if (type.isPrimitive) {
					syntheticPrimitives[i] = when (type) {
						java.lang.Boolean.TYPE -> false
						java.lang.Integer.TYPE -> 0
						java.lang.Long.TYPE -> 0L
						java.lang.Short.TYPE -> (0).toShort()
						java.lang.Byte.TYPE -> (0).toByte()
						java.lang.Character.TYPE -> (0).toChar()
						java.lang.Float.TYPE -> 0f
						java.lang.Double.TYPE -> 0.0
						else -> throw IllegalArgumentException("mamma mia: ${type.name}")
					}
				}
			}
		}
	}

	override fun write(out: JsonWriter, value: T) {
		out.beginObject()

		for ((field, adapter) in types) {
			out.name(field.name)
			@Suppress("unchecked_cast")
			(adapter as TypeAdapter<Any>).write(out, (field as KProperty1<T, Any>).get(value))
		}

		out.endObject()
	}

	override fun read(reader: JsonReader): T {
		// таблица присутствия значений (если значение true то на i было значение внутри json)
		val presentValues = BooleanArray(types.size + (if (storesJson) 1 else 0))
		val readValues = arrayOfNulls<Any>(types.size + (if (storesJson) 1 else 0))

		if (storesJson)
			presentValues[presentValues.size - 1] = true

		@Suppress("name_shadowing")
		var reader = reader

		// Если нам необходимо читать объект как набор данных массива, то давай
		if (asJsonArray) {
			val iterator = types.iterator()
			var fieldId = 0

			if (storesJson) {
				val readArray = TypeAdapters.JSON_ELEMENT.read(reader)

				if (readArray !is JsonArray) {
					throw JsonParseException("Expected JSON element to be an Array, ${readArray::class.qualifiedName} given")
				}

				reader = JsonTreeReader(readArray)
				readValues[readValues.size - 1] = enrollList(flattenJsonElement(readArray) as List<Any>, Starbound.STRING_INTERNER::intern)
			}

			reader.beginArray()

			while (reader.peek() != JsonToken.END_ARRAY) {
				if (!iterator.hasNext()) {
					val name = fieldId.toString()

					if (!storesJson && loggedMisses.add(name)) {
						if (currentSymbolicName == null) {
							LOGGER.warn("${bound.qualifiedName} has no property for storing $name")
						} else {
							LOGGER.warn("${bound.qualifiedName} has no property for storing $name (reading: $currentSymbolicName)")
						}
					}

					reader.skipValue()
					fieldId++

					continue
				}

				val tuple = iterator.next()
				val (field, adapter) = tuple

				try {
					readValues[fieldId] = (tuple.transformer as (Any?) -> Any?)(adapter.read(reader))
					presentValues[fieldId] = true
				} catch(err: Throwable) {
					throw JsonSyntaxException("Reading field ${field.name} for ${bound.qualifiedName}", err)
				}

				fieldId++
			}
		// иначе - читаем как json object
		} else {
			var json: JsonObject by Delegates.notNull()
			val hasFlatValues = types.any { it.isFlat }

			if (storesJson || hasFlatValues) {
				val readMap = TypeAdapters.JSON_ELEMENT.read(reader)

				if (readMap !is JsonObject) {
					throw JsonParseException("Expected JSON element to be a Map, ${readMap::class.qualifiedName} given")
				}

				json = readMap
				reader = JsonTreeReader(readMap)
				readValues[readValues.size - 1] = enrollMap(flattenJsonElement(readMap) as Map<String, Any>, Starbound.STRING_INTERNER::intern)
			}

			reader.beginObject()

			while (reader.peek() != JsonToken.END_OBJECT) {
				val name = reader.nextName()
				val fieldId = name2index.getInt(name)

				if (fieldId == -1) {
					if (!storesJson && !hasFlatValues && logMisses && loggedMisses.add(name)) {
						if (currentSymbolicName == null) {
							LOGGER.warn("${bound.qualifiedName} has no property for storing $name ")
						} else {
							LOGGER.warn("${bound.qualifiedName} has no property for storing $name (reading: $currentSymbolicName)")
						}
					}

					reader.skipValue()
				} else {
					val tuple = types[fieldId]

					if (tuple.isFlat) {
						reader.skipValue()
						continue
					}

					val (field, adapter) = tuple

					try {
						readValues[fieldId] = (tuple.transformer as (Any?) -> Any?)(adapter.read(reader))
						presentValues[fieldId] = true
					} catch(err: Throwable) {
						throw JsonSyntaxException("Reading field ${field.name} for ${bound.qualifiedName}", err)
					}
				}
			}

			for ((i, property) in types.withIndex()) {
				if (property.isFlat) {
					try {
						val read = property.adapter.read(JsonTreeReader(json))

						if (read != null) {
							presentValues[i] = true
							readValues[i] = read
						}
					} catch(err: Throwable) {
						throw JsonSyntaxException("Reading flat field ${property.property.name} for ${bound.qualifiedName}", err)
					}
				}
			}
		}

		if (asJsonArray) {
			reader.endArray()
		} else {
			reader.endObject()
		}

		// если у нас есть все значения для конструктора (все значения присутствуют в исходном json объекте)
		// или у нас нет синтетического конструктора, то делаем вызов "обычного" конструктора
		if (syntheticFactory == null || presentValues.all { it }) {
			for ((i, tuple) in types.withIndex()) {
				val (field) = tuple

				if (readValues[i] == null) {
					if (!tuple.returnType.isMarkedNullable) {
						throw JsonSyntaxException("Field ${field.name} of ${bound.qualifiedName} does not accept nulls")
					}

					if (!regularFactory.parameters[i].isOptional && !presentValues[i]) {
						throw JsonSyntaxException("Field ${field.name} of ${bound.qualifiedName} must be defined (even just as null)")
					}
				}
			}

			@Suppress("unchecked_cast")
			return regularFactory.call(*readValues as Array<out Any>)
		// иначе - в бой вступает синтетический конструктор, с флагами "значения по умолчанию"
		} else {
			// количество bitflag'ов значений по умолчанию
			val ints = if (presentValues.size % 31 == 0) presentValues.size / 31 else presentValues.size / 31 + 1
			// максимум считанных значений + bitflag аргументы значений по умолчанию + null типа DefaultConstructorMarker
			val copied = readValues.copyOf(readValues.size + ints + 1)
			var intIndex = readValues.size
			var target = 0
			var targetMove = 0

			for (bool in presentValues) {
				if (!bool) {
					target = target.or(1.shl(targetMove))
				}

				targetMove++

				if (targetMove >= 32) {
					copied[intIndex++] = target
					target = 0
					targetMove = 0
				}
			}

			if (targetMove != 0) {
				copied[intIndex] = target
			}

			val syntheticPrimitives = syntheticPrimitives!!

			for ((i, tuple) in types.withIndex()) {
				if (copied[i] != null) {
					continue
				}

				val (field) = tuple
				val param = regularFactory.parameters[i]

				if (!param.isOptional && !presentValues[i]) {
					throw JsonSyntaxException("Field ${field.name} of ${bound.qualifiedName} is missing")
				}

				if (tuple.returnType.isMarkedNullable) {
					continue
				}

				if (param.isOptional && !presentValues[i]) {
					copied[i] = syntheticPrimitives[i]
					continue
				}

				throw JsonSyntaxException("Field ${field.name} of ${bound.qualifiedName} does not accept nulls")
			}

			return syntheticFactory.newInstance(*copied)
		}
	}

	/**
	 * Позволяет построить класс [FactoryAdapter] на основе заданных параметров
	 */
	class Builder<T : Any>(val clazz: KClass<T>) {
		constructor(clazz: KClass<T>, vararg fields: KProperty1<T, *>) : this(clazz) {
			for (field in fields) {
				auto(field)
			}
		}

		private val types = ArrayList<PackedProperty<T, *>>()

		/**
		 * Принимает ли класс *последним* аргументом JSON структуру
		 *
		 * На самом деле, JSON "заворачивается" в [ImmutableMap], или [ImmutableList] если указано [asList]/[inputAsList]
		 *
		 * Поэтому, конструктор класса ОБЯЗАН принимать [Map]/[ImmutableMap] или [List]/[ImmutableList] первым аргументом,
		 * иначе поиск конструктора завершится неудчаей
		 */
		var storesJson = false

		/**
		 * Логировать ли несуществующие свойства у класса когда они попадаются в исходной JSON структуре
		 */
		var logMisses = true

		fun storesJson(flag: Boolean = true): Builder<T> {
			storesJson = flag
			return this
		}

		/**
		 * Логировать ли json значения которые нет в списке свойств
		 */
		fun logMisses(flag: Boolean = true): Builder<T> {
			logMisses = flag
			return this
		}

		/**
		 * Добавляет свойство с определённым [adapter]
		 */
		fun <V> add(field: KProperty1<T, V>, adapter: TypeAdapter<V>): Builder<T> {
			types.add(PackedProperty(field, adapter, isFlat = false))
			return this
		}

		/**
		 * Добавляет свойство с определённым [adapter], которое находится на том же уровне что и данный объект внутри JSON структуры
		 */
		fun <V> flat(field: KProperty1<T, V>, adapter: TypeAdapter<V>): Builder<T> {
			types.add(PackedProperty(field, adapter, isFlat = true))
			return this
		}

		/**
		 * Добавляет поля без generic типов и без преобразователей
		 */
		@Suppress("unchecked_cast")
		fun auto(vararg fields: KProperty1<T, *>): Builder<T> {
			for (field in fields)
				auto(field)

			return this
		}

		/**
		 * Добавляет поля без generic типов и без преобразователей
		 */
		@Suppress("unchecked_cast")
		fun autoFlat(vararg fields: KProperty1<T, *>): Builder<T> {
			for (field in fields)
				autoFlat(field)

			return this
		}

		/**
		 * Автоматически определяет тип поля и необходимый адаптор типа к нему
		 */
		@Suppress("unchecked_cast")
		fun <In> auto(field: KProperty1<T, In>, transformer: (In) -> In = { it }): Builder<T> {
			val returnType = field.returnType
			val classifier = returnType.classifier as? KClass<*> ?: throw ClassCastException("Unable to cast ${returnType.classifier} to KClass of property ${field.name}!")

			if (classifier.isSubclassOf(List::class)) {
				throw IllegalArgumentException("${field.name} is a List, please use autoList() method instead")
			}

			if (classifier.isSubclassOf(Map::class)) {
				throw IllegalArgumentException("${field.name} is a Map, please use autoMap() method instead")
			}

			types.add(PackedProperty(field, LazyTypeProvider(classifier.java) as TypeAdapter<Any?>, transformer = transformer as (Any?) -> Any?, isFlat = false))
			return this
		}

		/**
		 * Автоматически определяет тип поля и необходимый адаптор типа к нему
		 */
		@Suppress("unchecked_cast")
		fun <In> autoFlat(field: KProperty1<T, In>, transformer: (In) -> In = { it }): Builder<T> {
			val returnType = field.returnType
			val classifier = returnType.classifier as? KClass<*> ?: throw ClassCastException("Unable to cast ${returnType.classifier} to KClass of property ${field.name}!")

			if (classifier.isSubclassOf(List::class)) {
				throw IllegalArgumentException("${field.name} is a List")
			}

			if (classifier.isSubclassOf(Map::class)) {
				throw IllegalArgumentException("${field.name} is a Map")
			}

			types.add(PackedProperty(field, LazyTypeProvider(classifier.java) as TypeAdapter<Any?>, transformer = transformer as (Any?) -> Any?, isFlat = true))
			return this
		}

		/**
		 * Добавляет поле, которое содержит список значений V (без null).
		 *
		 * Список неизменяем (создаётся объект [ImmutableList])
		 */
		fun <V : Any> list(field: KProperty1<T, List<V>?>, type: Class<V>, transformer: (V) -> V = { it }): Builder<T> {
			types.add(PackedProperty(field, ListAdapter(type, transformer).nullSafe(), isFlat = false))
			return this
		}

		/**
		 * Добавляет поле, которое содержит список значений V (без null).
		 *
		 * Список неизменяем (создаётся объект [ImmutableList])
		 */
		inline fun <reified V : Any> autoList(field: KProperty1<T, List<V>?>, noinline transformer: (V) -> V = { it }): Builder<T> {
			return add(field, ListAdapter(V::class.java, transformer).nullSafe())
		}

		/**
		 * Добавляет поле-таблицу, которое кодируется как [[key, value], [key, value], ...]
		 *
		 * Таблица неизменяема (создаётся объект [ImmutableMap])
		 */
		fun <K, V> mapAsArray(field: KProperty1<T, Map<K, V>>, keyType: Class<K>, valueType: Class<V>): Builder<T> {
			types.add(PackedProperty(field, MapAdapter(keyType, valueType), isFlat = false))
			return this
		}

		/**
		 * Добавляет поле-таблицу, кодирование зависит от контекста:
		 * * Если [K] является [String], кодирование происходит как Json Object
		 * * Иначе кодирование происходит как Json Array
		 *
		 * Таблица неизменяема (создаётся объект [ImmutableMap])
		 */
		inline fun <reified K, reified V> autoMap(field: KProperty1<T, Map<K, V>>): Builder<T> {
			if (K::class == String::class)
				return this.mapAsObject(field as KProperty1<T, Map<String, V>?>, V::class.java)

			return this.mapAsArray(field, K::class.java, V::class.java)
		}

		/**
		 * Добавляет поле-таблицу, которое кодируется как [[key, value], [key, value], ...]
		 *
		 * Таблица неизменяема (создаётся объект [ImmutableMap])
		 */
		fun <K : Any, V : Any> mapAsArray(field: KProperty1<T, Map<K, V>?>, keyType: KClass<K>, valueType: KClass<V>): Builder<T> {
			types.add(PackedProperty(field, MapAdapter(keyType.java, valueType.java).nullSafe(), isFlat = false))
			return this
		}

		/**
		 * Добавляет поле-таблицу, которое кодируется как {"a": value, "b": value, ...}
		 *
		 * Таблица неизменяема (создаётся объект [ImmutableMap])
		 */
		fun <V> mapAsObject(field: KProperty1<T, Map<String, V>?>, valueType: Class<V>): Builder<T> {
			types.add(PackedProperty(field, String2ObjectAdapter(valueType).nullSafe(), isFlat = false))
			return this
		}

		/**
		 * Добавляет поле-таблицу, которое кодируется как {"a": value, "b": value, ...}
		 *
		 * Таблица неизменяема (создаётся объект [ImmutableMap])
		 */
		fun <V : Any> mapAsObject(field: KProperty1<T, Map<String, V>?>, valueType: KClass<V>): Builder<T> {
			types.add(PackedProperty(field, String2ObjectAdapter(valueType.java).nullSafe(), isFlat = false))
			return this
		}

		var asList = false

		fun inputAsMap(): Builder<T> {
			asList = false
			return this
		}

		fun inputAsList(): Builder<T> {
			asList = true
			return this
		}

		fun build(): FactoryAdapter<T> {
			check(!asList || types.none { it.isFlat }) { "Can't have both flat properties and json data array layout" }

			return FactoryAdapter(
				bound = clazz,
				types = ImmutableList.copyOf(types),
				asJsonArray = asList,
				storesJson = storesJson,
				logMisses = logMisses,
			)
		}
	}

	companion object {
		private val LOGGER = LogManager.getLogger()

		var currentSymbolicName by ThreadLocal<String>()
	}
}
