package ru.dbotthepony.kstarbound.world.entities

import ru.dbotthepony.kbox2d.api.*
import ru.dbotthepony.kbox2d.dynamics.contact.AbstractContact
import ru.dbotthepony.kvector.util2d.AABB
import ru.dbotthepony.kvector.vector.ndouble.Vector2d

enum class CollisionResolution {
	STOP,
	BOUNCE,
	PUSH,
	SLIDE,
}

abstract class MovementController<T : IEntity>(val entity: T) : IContactListener {
	val world = entity.world
	open var position by entity::position
	open var angle by entity::angle

	/**
	 * Уничтожает данный movement controller.
	 *
	 * Вызывается изнутри [Entity.remove]
	 */
	open fun destroy() {
		body.world.destroyBody(body)
	}

	private val bodyInit = lazy {
		world.physics.createBody(BodyDef(
			position = position,
			angle = angle,
			type = BodyType.DYNAMIC,
			userData = this
		))
	}

	/**
	 * Было ли создано физическое тело
	 */
	val bodyInitialized: Boolean
		get() = bodyInit.isInitialized()

	/**
	 * Физическое тело данного контроллера перемещения
	 *
	 * Не создаётся в мире пока к нему не обратятся
	 */
	protected val body by bodyInit

	/**
	 * Вызывается изнутри [Entity], когда оно спавнится в самом мире.
	 *
	 * Так как никто не запрещает нам создавать физические тела в физическом мире
	 * до того, как мы появимся в самом мире, это негативно сказывается на производительности И не является корректным
	 * поведением.
	 *
	 * Поэтому, прикреплять фигуры к физическому телу лучше всего из данной функции.
	 */
	open fun onSpawnedInWorld() {

	}

	open val velocity get() = body.linearVelocity

	open fun applyVelocity(velocity: Vector2d) {
		if (velocity != Vector2d.ZERO) {
			body.applyLinearImpulseToCenter(velocity)
		}
	}

	/**
	 * Проверяет, находится ли что-либо под нами
	 */
	open val onGround: Boolean get() {
		for (contact in body.contactEdgeIterator) {
			if (contact.contact.manifold.localNormal.dot(world.gravity.normalized) >= 0.97) {
				return true
			}
		}

		return false
	}

	/**
	 * World space AABB, by default returning combined AABB of physical body.
	 */
	open val worldAABB: AABB get() {
		return body.worldSpaceAABB
	}

	/**
	 * This is called on each world step to update variables and account of changes of
	 * physics world and this physics body.
	 */
	open fun think(delta: Double) {
		mutePositionChanged = true
		position = body.position
		angle = body.angle
		mutePositionChanged = false
	}

	protected open fun onTouchSurface(velocity: Vector2d, normal: Vector2d) {
		entity.onTouchSurface(velocity, normal)
	}

	protected var mutePositionChanged = false

	open fun notifyPositionChanged() {
		if (mutePositionChanged) {
			return
		}

		body.setTransform(entity.position, entity.angle)
	}
}

/**
 * Movement controller for logical entities, which does nothing.
 */
class LogicalMovementController(entity: Entity) : MovementController<Entity>(entity) {
	override val worldAABB: AABB get() = AABB(position, position)

	// Dummies never touch anything, since they don't have Box2D body
	override val onGround: Boolean = false
	override val velocity: Vector2d = Vector2d.ZERO

	override fun think(delta: Double) {
		// no-op
	}

	override fun notifyPositionChanged() {
		// no-op
	}

	override fun beginContact(contact: AbstractContact) {
		// no-op
	}

	override fun endContact(contact: AbstractContact) {
		// no-op
	}

	override fun preSolve(contact: AbstractContact, oldManifold: Manifold) {
		// no-op
	}

	override fun postSolve(contact: AbstractContact, impulse: ContactImpulse) {
		// no-op
	}

	companion object {
		private val DUMMY_AABB = AABB.rectangle(Vector2d.ZERO, 0.1, 0.1)
	}
}
